import json

from requests.exceptions import RequestException


class BaseException(Exception):
    def __init__(self, message=None, error_code=None, *args, **kwargs):
        if error_code:
            assert isinstance(error_code, str), "message should be str"
            assert error_code.isupper(), "error_code should be uppercase - SOMETHING_FAILED"

        if message:
            assert isinstance(message, str), "message should be str"
            assert message, "message should be populated"

        self.error_code = error_code or self.__class__.DEFAULT_ERROR_CODE
        self.message = message or self.__class__.DEFAULT_MESSAGE

        super().__init__(*args, **kwargs)

    def __str__(self):
        self.err = {"message": self.message, "error_code": self.error_code}

        return json.dumps(self.err)


class UnexpectedException(BaseException):
    DEFAULT_ERROR_CODE = "INTERNAL_ERROR"
    DEFAULT_MESSAGE = "An unexpected error occurred. We are working on resolving the issue."


class InternalServerError(UnexpectedException):
    pass


class ValidationException(BaseException, TypeError, ValueError):
    DEFAULT_ERROR_CODE = "VALIDATION_ERROR"
    DEFAULT_MESSAGE = "Unexpected validation error."


class UnsupportedEnvException(ValidationException):
    DEFAULT_ERROR_CODE = "UNSUPPORTED_ENV_ERROR"
    DEFAULT_MESSAGE = "An unsupported environment exception occurred."


class ResourceAlreadyExists(ValidationException):
    DEFAULT_ERROR_CODE = "RESOURCE_ALREADY_EXISTS"
    DEFAULT_MESSAGE = "Resource already exists."


class ResourceNotFound(ValidationException):
    DEFAULT_ERROR_CODE = "RESOURCE_NOT_FOUND"
    DEFAULT_MESSAGE = "Resource not found."


# Can we adjust this exception to include url and then fire a hook? Would save us time
class ExternalServiceException(BaseException, RequestException):
    DEFAULT_ERROR_CODE = "EXTERNAL_SERVICE_FAIL"
    DEFAULT_MESSAGE = "Call to underlying service failed."

    def __str__(self):
        err = {"message": self.message, "error_code": self.error_code}

        if self.request is not None:
            err.update({"method": self.request.method, "url": self.request.url})

        if self.response is not None:
            err.update(
                {
                    "reason": self.response.reason,
                    "status_code": self.response.status_code,
                    "response": self.response.text,
                }
            )

        return json.dumps(err)


class WebhookServiceException(BaseException):
    DEFAULT_ERROR_CODE = "GOOGLE_SERVICE_FAIL"
    DEFAULT_MESSAGE = "Failed to call Google Webhook Service"

    def __init__(self, url: str, status_code: int, body: str, message=None, error_code=None):
        self.url = url
        self.status_code = status_code
        self.body = body
        super().__init__(message, error_code)

    def __str__(self):
        err = {
            "message": self.message,
            "error_code": self.error_code,
            "status_code": self.status_code,
            "url": self.url,
            "body": self.body,
        }

        return json.dumps(err)
