from static.config import DELTA_SORT
from static.utils import Log
from cbc import CSR

logger = Log()


def set_core_features(event: dict) -> None:
    """Function to orchestrate all feature provisioning
    Each feature from the delta is pointed the corresponding provisioning function here
    """
    final_flags = {}
    # Feature flag based features will be added automatically
    # The map below only contains features that need backend api calls
    # occasionally flags are returned from these funcs that need to be overrode
    feature_functions = {
        "psc:feature:defense": {
            "CREATE": create_defense,
            "DELETE": delete_defense,
        },
        "psc:feature:vulnerability": {
            "CREATE": create_vulnerability,
            "DELETE": delete_vulnerability,
        },
        "psc:feature:vulnerabilityendpoint": {
            "CREATE": create_vulnerabilityendpoint,
            "DELETE": delete_vulnerabilityendpoint,
        },
        "psc:feature:prevention": {
            "CREATE": create_prevention,
            "DELETE": delete_prevention,
        },
        "psc:feature:threathunter": {
            "CREATE": create_threathunter,
            "DELETE": delete_threathunter,
        },
        "psc:feature:retention": {
            "CREATE": create_retention,
            "DELETE": delete_retention,
        },
        "psc:feature:publiccloud": {
            "CREATE": create_publiccloud,
            "DELETE": delete_publiccloud,
        },
        "psc:feature:xdr": {
            "CREATE": create_xdr,
            "DELETE": delete_xdr,
        },
        "psc:feature:ransomwarerecovery": {
            "DELETE": delete_ransomwarerecovery,
        },
        "psc:feature:hbfw": {
            "CREATE": create_hbfw,
            "DELETE": delete_hbfw,
        },
        "psc:feature:cisbenchmark": {
            "CREATE": create_cisbenchmark,
            "DELETE": delete_cisbenchmark,
        },
    }

    csr = CSR(event["request_data"]["environment"])
    feature_catalog = csr.get_features()

    org_id = event.get("final_state", {}).get("id") or event.get("initial_state", {}).get("id")
    org_key = event.get("final_state", {}).get("orgKey") or event.get("initial_state", {}).get("orgKey")
    sorted_delta = dict(sorted(event["final_state"]["delta"].items(), key=lambda tup: DELTA_SORT[tup[1]["change"]]))

    for feature_id, feature_delta in sorted_delta.items():  # for each feature in the feature delta
        feature_context = {
            "org_id": org_id,
            "org_key": org_key,
            "feature": feature_id,
            "config": feature_delta,
            "calculated_features": event["final_state"]["calculated_features"],
            "change": feature_delta["change"],
            "feature_config": next((item for item in list(feature_catalog) if item["feature_id"] == feature_id), None),
        }

        # If feature needs to set flags
        if feature_context["feature_config"]:
            initial_flags = build_feature_flags(feature_context["feature_config"], feature_context["change"])
            final_flags = {**final_flags, **initial_flags}

        # If feature needs backend config
        if feature_context["change"] in feature_functions.get(feature_id, {}):
            if "normalize" not in feature_context["config"]["custom_attr"] or (
                feature_context["feature"] in ["psc:feature:defense", "psc:feature:ransomwarerecovery"]
                and "normalize" in feature_context["config"]["custom_attr"]
            ):
                # Occasionally flags are overridden if feature flags are shared between features (Yikes!)
                override_flags = feature_functions[feature_id][feature_context["change"]](csr, feature_context)
                logger.send_info({"override flags": override_flags})
                final_flags = {**final_flags, **override_flags}

        logger.send_info({"feature_context": feature_context, "final_flags": final_flags})

    # Gather all flags and set them once (Avoid replication issues)
    if final_flags:
        csr.post_org_feature_flags(str(org_id), final_flags)


def normalize_org_features(event: dict) -> None:
    """Function that disables all products on a given organization
    This the function used to confirm clean slate on newly-created orgs
    """
    csr = CSR(event["request_data"]["environment"])

    delete_event = dict(event)
    features = csr.get_features()  # repetitive
    delta = {}

    for ft in features:
        delta[ft["feature_id"]] = {"licenses": 0, "change": "DELETE", "custom_attr": {"normalize": True}}

    delete_event["final_state"]["calculated_features"] = []
    delete_event["final_state"]["delta"] = delta

    return set_core_features(delete_event)


def revert_org_features(event: dict) -> None:
    """Function that reverts all products that failed to provision on an org
    This the function used to make sure no orgs have partial features enabled
    """

    delta = {}
    for feature, value in event["final_state"]["delta"].items():
        delta[feature] = value

        if value["change"] == "CREATE":
            delta[feature]["change"] = "DELETE"
            delta[feature]["custom_attr"] = {"revert": True}
        elif value["change"] == "DELETE":
            delta[feature]["change"] = "CREATE"
            delta[feature]["custom_attr"] = {"revert": True}

    revert_event = dict(event)
    revert_event["final_state"]["delta"] = delta

    return set_core_features(revert_event)


def reentitle_all_features(event: dict) -> None:
    """Function that enables all features on a given organization"""
    delta = event["final_state"]["delta"]

    csr = CSR(event["request_data"]["environment"])
    features = csr.get_features()  # repetitive

    for key, value in delta.items():
        value["change"] = "CREATE"

    for feature in features:  # Delete all features that are not currently enabled
        if feature["feature_id"] not in delta.keys():
            delta[feature["feature_id"]] = {
                "licenses": 0,
                "change": "DELETE",
                "custom_attr": {},
                "external_name": feature["external_name"],
            }

    logger.send_info(f"REENTITLE ALL DELTA: {delta}")

    return set_core_features(event)


def reentitle_specific_features(event: dict) -> None:
    """Function that enables specific features on a given organization"""

    requested_entitlement = event["request_data"]["entitlements"][0]
    initial_delta = dict(event["final_state"]["delta"])
    delta = event["final_state"]["delta"]

    csr = CSR(event["request_data"]["environment"])
    catalog = csr.get_entitlement_catalog()

    logger.send_info(f"INITIAL DELTA: {initial_delta}")

    try:
        mapped_catalog = dict((f"{p['system']}:{p['product_id']}", p) for i, p in enumerate(catalog))
        product = mapped_catalog[f"{requested_entitlement['source']}:{requested_entitlement['product']}"]
        product_features = product["features"]

        for i, v in delta.items():
            v["change"] = "NONE"

        for feature in product_features:
            delta[feature["feature_id"]]["change"] = "CREATE"

    except Exception as e:
        logger.send_warning(
            f"Failed to build delta for product {requested_entitlement['product']} for system {requested_entitlement['source']}."
        )
        logger.send_exception(str(e))
    else:
        logger.send_info(f"REENTITLE ONE DELTA: {delta}")
        set_core_features(event)

    return


def override_feature_config(data: dict, dependencies: list) -> bool:
    """This is used to determine co dependencies withing features"""

    if [feature for feature in data["calculated_features"] if feature["feature_id"] in dependencies]:
        logger.send_info(f"{data['feature']} is overriding flags: {str(dependencies)}")
        return True

    return False


def build_feature_flags(feature: dict, change: str) -> dict:
    """Function that builds the feature flags for each feature passed in"""
    return dict.fromkeys(feature["config"]["flags"], False if change == "DELETE" else True)


def create_threathunter(csr: CSR, data: dict) -> dict:
    """Function to provision threathunter and all it's steps"""

    csr.post_org_policyservice_enable_threathunter(data["org_key"])

    return {}


def create_prevention(csr: CSR, data: dict) -> dict:
    """Function to provision prevention and all it's steps"""

    csr.post_org_defense_config(data["org_id"], False)
    csr.post_org_prevention_policy(data["org_id"])

    return {}


def create_defense(csr: CSR, data: dict) -> dict:
    """Function to provision defense and all it's steps"""

    csr.post_org_defense_config(data["org_id"], False)

    return {}


def create_vulnerability(csr: CSR, data: dict) -> dict:
    """Function to provision all steps for vulnerability"""

    csr.put_org_vulnerability_assessment(
        data["org_key"],
        [
            {"name": "WORKLOAD", "enabled": True},
        ],
    )

    return {}


def create_publiccloud(csr: CSR, data: dict) -> dict:
    """Function to provision publiccloud and all of it's config"""

    csr.put_org_vulnerability_assessment(
        data["org_key"],
        [
            {"name": "AWS", "enabled": True},
            {"name": "AZURE", "enabled": True},
            {"name": "GCP", "enabled": True},
        ],
    )

    return {}


def create_vulnerabilityendpoint(csr: CSR, data: dict) -> dict:
    """Function to provision all steps for vulnerability endpoint"""

    csr.put_org_vulnerability_assessment(
        data["org_key"],
        [
            {"name": "WORKLOAD", "enabled": True},
            {"name": "ENDPOINT", "enabled": True},
            {"name": "AWS", "enabled": True},
            {"name": "AZURE", "enabled": True},
            {"name": "GCP", "enabled": True},
        ],
    )

    return {}


def create_retention(csr: CSR, data: dict) -> dict:
    """Function to provision retention and it's config"""

    csr.put_org_retention_class(data["org_key"], {"retention_days": data["config"]["custom_attr"]["retention_days"]})

    return {}


def create_xdr(csr: CSR, data: dict) -> dict:
    """Function to provision xdr and all of it's config"""

    csr.put_org_policyservice_features(
        data["org_key"],
        [
            {"name": data["feature"], "enabled": True},
        ],
    )

    return {}


def create_hbfw(csr: CSR, data: dict) -> dict:
    """Function to provision hbfw and all of it's config"""

    csr.post_org_global_sensor_settings(data["org_id"], "true")  # toggle global sensors on

    csr.post_org_sensor_settings(
        data["org_key"],
        csr.build_sensor_config("CbFirewallEnabled", "<OnlyIf:SensorVersion:Atleast:3.9.0.0>", data["feature"], True),
    )  # set sensor settings

    return {}


def create_cisbenchmark(csr: CSR, data: dict) -> dict:
    """Function to provision all steps for cis benchmarks"""

    csr.put_org_compliance_assessment(
        data["org_key"],
        [
            {"deployment_type": "WORKLOAD", "enabled": True},
        ],
    )

    return {}


def delete_hbfw(csr: CSR, data: dict) -> dict:
    """Function to deprovision hbfw and all of it's config"""

    csr.post_org_sensor_settings(
        data["org_key"],
        csr.build_sensor_config("CbFirewallEnabled", "<OnlyIf:SensorVersion:Atleast:3.9.0.0>", data["feature"], False),
    )  # set sensor settings

    return {}


def delete_ransomwarerecovery(csr: CSR, data: dict) -> dict:
    """Function to deprovision ransomewarerecovery and all it's steps"""

    return {"conferAV": "default"}


def delete_xdr(csr: CSR, data: dict) -> dict:
    """Function to deprovision xdr and all of it's config"""

    csr.put_org_policyservice_features(data["org_key"], [{"name": data["feature"], "enabled": False}])

    return {}


def delete_threathunter(csr: CSR, data: dict) -> dict:
    """Function to deprovision threathunter and all it's steps"""

    csr.post_org_policyservice_disable_threathunter(data["org_key"])

    # Should we still be deleting tags and alerts? Does TOOLS-2539 change this functionality?
    if csr.get_org_watchlists(data["org_key"]):
        csr.delete_org_watchlist_tags(data["org_key"])
        csr.delete_org_watchlist_alerts(data["org_key"])

    return {}


def delete_prevention(csr: CSR, data: dict) -> dict:
    """Function to deprovision ngav and all it's steps"""

    override_bool = override_feature_config(data, ["psc:feature:defense"])

    if override_bool:
        csr.post_org_prevention_upgrade(data["org_id"])
    else:
        csr.post_org_prevention_downgrade(data["org_id"])

    return {"cbDefense": override_bool}


def delete_defense(csr: CSR, data: dict) -> dict:
    """Function to deprovision defense and all it's steps"""

    override_bool = override_feature_config(data, ["psc:feature:prevention"])

    csr.post_org_defense_config(data["org_id"], (not override_bool))

    return {"cbDefense": override_bool}


def delete_vulnerability(csr: CSR, data: dict) -> dict:
    """Function to deprovision all steps for vulnerability"""

    override_flag_bool = override_feature_config(data, ["psc:feature:vulnerabilityendpoint", "psc:feature:publiccloud"])
    override_vw_bool = override_feature_config(data, ["psc:feature:vulnerabilityendpoint"])

    csr.put_org_vulnerability_assessment(
        data["org_key"],
        [
            {"name": "WORKLOAD", "enabled": override_vw_bool},
        ],
    )

    return {"vulnerabilityAssessment": override_flag_bool}


def delete_publiccloud(csr: CSR, data: dict) -> dict:
    """Function to deprovision publicclud and it's config"""

    override_flag_bool = override_feature_config(data, ["psc:feature:vulnerabilityendpoint", "psc:feature:vulnerability"])
    override_pc_bool = override_feature_config(data, ["psc:feature:vulnerabilityendpoint"])

    csr.put_org_vulnerability_assessment(
        data["org_key"],
        [
            {"name": "AWS", "enabled": override_pc_bool},
            {"name": "AZURE", "enabled": override_pc_bool},
            {"name": "GCP", "enabled": override_pc_bool},
        ],
    )

    return {"vulnerabilityAssessment": override_flag_bool}


def delete_vulnerabilityendpoint(csr: CSR, data: dict) -> dict:
    """Function to deprovision all steps for vulnerability endpoint"""

    override_flag_bool = override_feature_config(data, ["psc:feature:publiccloud", "psc:feature:vulnerability"])
    override_pc_bool = override_feature_config(data, ["psc:feature:publiccloud"])
    override_vw_bool = override_feature_config(data, ["psc:feature:vulnerability"])

    csr.put_org_vulnerability_assessment(
        data["org_key"],
        [
            {"name": "WORKLOAD", "enabled": override_vw_bool},
            {"name": "ENDPOINT", "enabled": False},
            {"name": "AWS", "enabled": override_pc_bool},
            {"name": "AZURE", "enabled": override_pc_bool},
            {"name": "GCP", "enabled": override_pc_bool},
        ],
    )

    return {"vulnerabilityAssessment": override_flag_bool}


def delete_retention(csr: CSR, data: dict) -> dict:
    """Function to deprovision retention and it's config"""

    csr.delete_org_retention_class(data["org_key"])

    return {}


def delete_cisbenchmark(csr: CSR, data: dict) -> dict:
    """Function to deprovision all steps for cis benchmarks"""

    csr.put_org_compliance_assessment(
        data["org_key"],
        [
            {"deployment_type": "WORKLOAD", "enabled": False},
        ],
    )

    return {}
