import psc_orgkey
from dataclasses import asdict, dataclass, field
from static.config import ENVIRONMENT, VCDR_PARENT
from os import environ


@dataclass
class Execution:
    start_time: str
    execution_urn: str
    request_urn: str
    execution_status: str
    region: str
    operation: str
    environment: str
    origin: str
    source: str
    external_urn: str
    request_data: dict = field(default_factory=dict)
    initial_state: dict = field(default_factory=dict)
    final_state: dict = field(default_factory=dict)
    error_info: dict = field(default_factory=dict)

    def inject_csp_defaults(self):
        if self.request_data.get("custom") is None:
            self.request_data["custom"] = {"is_csp": False}

        if self.environment not in ENVIRONMENT.__members__.keys():
            self.request_data["custom"]["is_csp"] = True

        if self.request_data.get("custom", {}).get("uuid") or self.request_data.get("org_uuid"):
            self.request_data["custom"]["is_csp"] = True

        if self.request_data.get("custom", {}).get("is_csp") is not True:
            self.request_data["custom"]["is_csp"] = False
        else:
            if environ.get("CSP_SERVICE_INSTANCE_ID"):
                self.request_data["custom"]["service_instance_id"] = environ["CSP_SERVICE_INSTANCE_ID"]

    def inject_origin_defaults(self):

        if self.operation == "REGISTER":
            if self.request_data.get("custom", {}).get("uuid"):
                self.final_state["csp_organization"] = {"id": self.request_data["custom"]["uuid"]}
        else:
            self.request_data["org_id"] = (psc_orgkey.decode_orgkey(orgkey=self.request_data["org_key"])).org_id

        if self.operation == "REENTITLE":
            self.request_data["custom"]["reentitle"] = "ALL" if len(self.request_data["entitlements"]) == 0 else "SINGLE"

        if not self.operation == "PURGE":
            for entitlement in self.request_data.get("entitlements", []):
                if entitlement.get("config") is None:
                    entitlement["config"] = {}

    def inject_vcdr_parent(self):
        if self.origin == "VCDR" and self.operation == "REGISTER" and self.request_data.get("custom", {}).get("is_csp") is False:
            self.request_data["parent"] = VCDR_PARENT.get(self.environment, "")

    def strip_sensitive_data(self):
        if not self.origin == "LOCAL":
            self.request_data.pop("user_token", None)

    def __post_init__(self):
        self.inject_origin_defaults()
        self.inject_csp_defaults()
        self.inject_vcdr_parent()
        self.strip_sensitive_data()

    def asdict(self):

        result = asdict(self)
        result["error-info"] = result["error_info"]
        result.pop("error_info", None)

        return result


def new_execution(event: dict) -> Execution:
    execution = Execution(
        start_time=event["start_time"],
        execution_urn=event["execution_urn"],
        request_urn=event["ticket"]["request_urn"],
        execution_status="IN_PROGRESS",
        region=event["ticket"]["request_data"]["region"],
        operation=event["ticket"]["request_data"]["operation"],
        environment=event["ticket"]["request_data"]["environment"],
        origin=event["ticket"]["request_data"]["origin"],
        source=event["ticket"]["request_data"]["source"],
        external_urn=f"{event['ticket']['request_data']['origin']}:{event['ticket']['request_data']['external_urn']}",
        request_data=event["ticket"]["request_data"],
        initial_state={},
        final_state={},
        error_info={},
    )
    return execution
