import traceback
from os import environ

from csp import csp
from cbc import CSR
from requests.exceptions import RequestException
from urllib3 import disable_warnings
from urllib3 import exceptions as url_exceptions

from static.utils import get_org_data_from_event, Log, set_os_variable

from entitlement import (
    get_calculated_entitlements,
    get_calculated_features,
    get_entitlement_delta,
    calculate_effective_feature_delta,
    get_org_feature_records,
    update_entitlement_execution,
    update_entitlement_record,
)
from exceptions import ExternalServiceException, InternalServerError, UnexpectedException, ValidationException
from feature import (
    normalize_org_features,
    set_core_features,
    reentitle_all_features,
    reentitle_specific_features,
    revert_org_features,
)
from organization import (
    create_organization,
    deregister_organization,
    get_org,
    get_flags,
    provision_admin_account,
    start_organization_contract,
)
from ticket import build_ticket, finalize_ticket_attributes, merge_tickets, parse_ticket_error, publish_csp, publish_sns
from validate import (
    validate_added_entitlements,
    validate_entitlement_dependencies,
    validate_existing_entitlements,
    validate_operation,
    validate_organization,
)
from hooks.chat import Webhook

logger = Log()

disable_warnings(url_exceptions.InsecureRequestWarning)


class DefaultEventHandler:
    def __init__(self, event: dict):
        # Set execution context
        self.state = str(event["execution_context"]["State"]["Name"]).casefold()
        self.urn = event["execution_context"]["Execution"]["Name"].replace("_", ":")
        # set event data
        self.event = event
        self.org = get_org_data_from_event(self.event["ticket"] if "ticket" in event.keys() else self.event)
        # set variables for logger
        set_os_variable("REQUEST_URN", event["execution_context"]["Execution"]["Name"].replace("_", ":"))
        set_os_variable("STATE_NAME", str(event["execution_context"]["State"]["Name"]).casefold())
        logger.send_info(f"STARTING STEP FUNCTION STATE: {event['execution_context']['State']['Name'].upper()}")

    def handle(self):
        raise InternalServerError(f"Lambda missing state {self.state}, internal error.")


class BuildStateMachineTicketHandler(DefaultEventHandler):
    def handle(self) -> dict:
        return build_ticket(self.event)


class ValidateIncomingOrganizationHandler(DefaultEventHandler):
    def handle(self) -> dict:
        validate_organization(self.event["ticket"])
        return self.event["ticket"]


class QueryOrganizationHandler(DefaultEventHandler):
    def handle(self) -> dict:
        return get_org(self.event)


class GetOrganizationFeatureDataHandler(DefaultEventHandler):
    def handle(self) -> dict:
        if not self.event.get("initial_state", {}).get("id", None):
            return {}

        return get_flags(self.event, org_only=True)


class GetGlobalFeatureDataHandler(DefaultEventHandler):
    def handle(self) -> dict:
        return get_flags(self.event, org_only=False)


class GetOrganizationFeaturesHandler(DefaultEventHandler):
    def handle(self) -> dict:
        return get_org_feature_records(self.event)


class MergeParallelTicketsHandler(DefaultEventHandler):
    def handle(self) -> dict:
        return merge_tickets(self.event)


class PreValidateOperationHandler(DefaultEventHandler):
    def handle(self) -> None:
        validate_operation(self.event)
        return


class PreValidateEntitlementsHandler(DefaultEventHandler):
    def handle(self) -> None:
        if self.event["ticket"]["operation"].casefold() in ["entitle", "register"]:
            validate_entitlement_dependencies(validate_added_entitlements(self.event["ticket"]))
        elif self.event["ticket"]["operation"].casefold() in ["reentitle"]:
            if self.event["ticket"]["request_data"]["custom"]["reentitle"].casefold() == "single":
                validate_existing_entitlements(self.event["ticket"])

        return


class PostValidateRegisterOperationHandler(DefaultEventHandler):
    def handle(self) -> None:
        validate_operation(self.event)
        return


class SendServiceInvitationHandler(DefaultEventHandler):
    def handle(self) -> None:
        csp_module = csp.CSP(self.urn, self.state)
        csp_module.create_service_invitation(
            {
                "TaskToken": self.event["execution_context"]["Task"]["Token"],
                "Region": environ["AWS_REGION"],
                "Environment": self.event["ticket"]["environment"].lower(),
                "ServiceInstanceId": self.event["ticket"]["request_data"]["custom"]["service_instance_id"],
            },
            self.event["ticket"]["request_data"]["contacts"][0]["first_name"],
            self.event["ticket"]["request_data"]["contacts"][0]["email"],
        )
        return


class EnableCSPServiceAccessHandler(DefaultEventHandler):
    def handle(self) -> None:
        csp_module = csp.CSP(self.urn, self.state)
        csp_module.enable_service_access(
            self.event["ticket"]["request_data"]["custom"]["service_instance_id"],
            self.event["ticket"]["final_state"]["csp_organization"]["id"],
        )
        csp_module.update_org_access(self.event["ticket"]["final_state"]["csp_organization"]["id"], "GRANT_ACCESS")
        csp_module.assign_service_instance_by_id(
            self.event["ticket"]["final_state"]["csp_organization"]["id"],
            self.event["ticket"]["request_data"]["custom"]["service_instance_id"],
        )
        return


class OverrideServiceInvitationHandler(DefaultEventHandler):
    def handle(self) -> dict:
        csp_module = csp.CSP(self.urn, self.state)
        if uuid := csp_module.create_org_using_user_context(
            self.event["ticket"]["request_data"]["org_name"],
            self.event["ticket"]["request_data"].pop("user_token"),
        ):
            csp_module.enable_service_access(self.event["ticket"]["request_data"]["custom"]["service_instance_id"], uuid)
            return {"id": uuid}
        else:
            raise InternalServerError("Unable to create csp org on behalf of user context.")


class CreateOrganizationHandler(DefaultEventHandler):
    def handle(self) -> dict:
        self.event["ticket"]["final_state"].update(create_organization(self.event["ticket"]))
        return self.event["ticket"]


class UpdateOrganizationHandler(DefaultEventHandler):
    def handle(self) -> None:
        start_organization_contract(self.event, self.event["initial_state"]["id"])
        return


class NormalizeFeaturesHandler(DefaultEventHandler):
    def handle(self) -> None:
        normalize_org_features(self.event)
        return


class RevertFeaturesHandler(DefaultEventHandler):
    def handle(self) -> None:
        calculated_entitlements = get_calculated_entitlements(self.event)
        self.event["ticket"]["final_state"]["calculated_features"] = get_calculated_features(self.event, calculated_entitlements)
        self.event["ticket"]["final_state"]["delta"] = calculate_effective_feature_delta(self.event)
        revert_org_features(self.event["ticket"])
        return


class AddSuperusersHandler(DefaultEventHandler):
    def handle(self) -> dict:
        if self.event.get("request_data").get("custom", {}).get("is_csp", False) is False:
            return provision_admin_account(self.event)
        return []


class AddFeaturesHandler(DefaultEventHandler):
    def handle(self) -> dict:
        calculated_entitlements = get_calculated_entitlements(self.event)
        self.event["ticket"]["final_state"]["calculated_features"] = get_calculated_features(self.event, calculated_entitlements)
        self.event["ticket"]["final_state"]["delta"] = calculate_effective_feature_delta(self.event)
        set_core_features(self.event["ticket"])
        return self.event["ticket"]


class ReEntitleAllEntitlementsHandler(DefaultEventHandler):
    def handle(self) -> dict:
        calculated_entitlements = get_calculated_entitlements(self.event)
        self.event["ticket"]["final_state"]["calculated_features"] = get_calculated_features(self.event, calculated_entitlements)
        self.event["ticket"]["final_state"]["delta"] = calculate_effective_feature_delta(self.event)
        reentitle_all_features(self.event["ticket"])
        return self.event["ticket"]


class ReEntitleSpecificEntitlementsHandler(DefaultEventHandler):
    def handle(self) -> dict:
        calculated_entitlements = get_calculated_entitlements(self.event)
        self.event["ticket"]["final_state"]["calculated_features"] = get_calculated_features(self.event, calculated_entitlements)
        self.event["ticket"]["final_state"]["delta"] = calculate_effective_feature_delta(self.event)
        reentitle_specific_features(self.event["ticket"])
        return self.event["ticket"]


class PurgeVCDROrganizationHandler(DefaultEventHandler):
    def handle(self) -> dict:

        if self.event["ticket"]["request_data"].get("org_uuid"):
            csp_module = csp.CSP(self.urn, self.state)

            org_instance_rsp = csp_module.get_org_service_instances(self.event["ticket"]["request_data"]["org_uuid"])

            # Only remove the service instance if the service id exists, else do nothing
            if self.event["ticket"]["request_data"]["custom"]["service_instance_id"] in org_instance_rsp.keys():
                csp_module.remove_service_instance_by_id(
                    self.event["ticket"]["request_data"]["org_uuid"],
                    self.event["ticket"]["request_data"]["custom"]["service_instance_id"],
                )
                org_instance_rsp.pop(self.event["ticket"]["request_data"]["custom"]["service_instance_id"], None)

            if len(org_instance_rsp.keys()) == 0:  # Only if no service instances are attached
                csp_module.update_org_access(self.event["ticket"]["request_data"]["org_uuid"], "DENY_ACCESS")

        deregister_organization(self.event["ticket"])  # ? May be better to have this in a separate state for deprov

        return self.event["ticket"]


class AddEntitlementsHandler(DefaultEventHandler):
    def handle(self) -> dict:
        calculated_entitlements = get_calculated_entitlements(self.event)
        self.event["ticket"]["final_state"]["entitlement_delta"] = get_entitlement_delta(calculated_entitlements, "NONE")
        return update_entitlement_record(self.event)


class GetEntitlementsDetailHandler(DefaultEventHandler):
    def handle(self) -> dict:
        if self.org["key"]:
            csr = CSR(self.event["ticket"]["request_data"]["environment"])
            result = csr.get_org_feature_details(self.org["key"])

            if self.event["ticket"]["operation"] == "REGISTER":
                self.event["ticket"]["final_state"]["stage"] = result.get("stage", None)
            else:
                self.event["ticket"]["initial_state"]["stage"] = result.get("stage", None)

        return self.event["ticket"]


class ParseStateMachineErrorHandler(DefaultEventHandler):
    def handle(self) -> dict:
        # Temporary solution, will have to move this to the event at a later point
        finalize_ticket_attributes(self.event)
        self.event["ticket"]["error-info"] = parse_ticket_error(dict(self.event["ticket"]["error-info"]))
        return self.event["ticket"]


class UpdateEntitlementsExecutionHandler(DefaultEventHandler):
    def handle(self) -> dict:
        finalize_ticket_attributes(self.event)
        return update_entitlement_execution(self.event)


class PublishSNSHandler(DefaultEventHandler):
    def handle(self) -> dict:
        publish_sns(self.event["ticket"])
        return self.event["ticket"]


class PublishCSPHandler(DefaultEventHandler):
    def handle(self) -> dict:
        publish_csp(self.event["ticket"])
        return self.event["ticket"]


class PublishWebhookFailureHandler(DefaultEventHandler):
    def handle(self) -> dict:

        if self.event["ticket"]["error-info"].get("Error", None) in [
            "ValidationException",
            "UnsupportedEnvException",
            "ResourceAlreadyExists",
            "ResourceNotFound",
            None,
        ]:
            return self.event["ticket"]

        if environ["GOOGLE_WEBHOOK"]:
            gchat = Webhook(
                environ["GOOGLE_WEBHOOK"],
                self.event["ticket"]["execution_urn"],
                self.org["name"],
                self.org["key"],
                self.org["env"],
                self.event["ticket"]["error-info"],
            )
            gchat.build()
            gchat.fire()

        return self.event["ticket"]


def event_handler(event: dict, context: dict) -> dict:
    """This function acts as the state machine front controller, each command is a block in the if statement"""
    default_handler = DefaultEventHandler(event)

    handlers = {
        "build state machine ticket": BuildStateMachineTicketHandler,
        "validate incoming organization": ValidateIncomingOrganizationHandler,
        "query organization": QueryOrganizationHandler,
        "get organization feature data": GetOrganizationFeatureDataHandler,
        "get global feature data": GetGlobalFeatureDataHandler,
        "get organization features": GetOrganizationFeaturesHandler,
        "merge parallel tickets": MergeParallelTicketsHandler,
        "pre validate operation": PreValidateOperationHandler,
        "pre validate entitlements": PreValidateEntitlementsHandler,
        "post validate register operation": PostValidateRegisterOperationHandler,
        "send service invitation": SendServiceInvitationHandler,
        "enable csp service access": EnableCSPServiceAccessHandler,
        "override service invitation": OverrideServiceInvitationHandler,
        "create organization": CreateOrganizationHandler,
        "update organization": UpdateOrganizationHandler,
        "normalize features": NormalizeFeaturesHandler,
        "revert features": RevertFeaturesHandler,
        "add superusers": AddSuperusersHandler,
        "add features": AddFeaturesHandler,
        "update features": AddFeaturesHandler,
        "reentitle all entitlements": ReEntitleAllEntitlementsHandler,
        "reentitle single entitlement": ReEntitleSpecificEntitlementsHandler,
        "purge vcdr organization": PurgeVCDROrganizationHandler,
        "add entitlements": AddEntitlementsHandler,
        "update entitlements": AddEntitlementsHandler,
        "get entitlements detail": GetEntitlementsDetailHandler,
        "parse error": ParseStateMachineErrorHandler,
        "update entitlements execution": UpdateEntitlementsExecutionHandler,
        "update entitlements execution failure": UpdateEntitlementsExecutionHandler,
        "publish sns": PublishSNSHandler,
        "publish sns failure": PublishSNSHandler,
        "publish csp": PublishCSPHandler,
        "publish csp failure": PublishCSPHandler,
        "publish webhook failure": PublishWebhookFailureHandler,
    }

    try:
        handler = handlers.get(default_handler.state, None)
        if handler is not None:
            return handler(event).handle()
        else:
            return default_handler.handle()
    except RequestException as e:
        raise ExternalServiceException(f"Call to underlying service failed: {str(e)}", request=e.request, response=e.response)
    except ValidationException as e:
        raise e
    except Exception as e:
        logger.send_exception("".join(traceback.TracebackException.from_exception(e).format()))
        logger.send_exception(e)

        raise UnexpectedException(str(e))
