import json
import boto3

from datetime import datetime
from os import environ
from typing import Union

from csp import csp
from cbc import CSR
from exceptions import InternalServerError
from static.utils import is_test_env, get_org_data_from_event, Log
from models.execution import new_execution

logger = Log()


def merge(dictionary1: dict, dictionary2: dict, path: list = None) -> dict:
    """merges two dictionaries recursively, updating dictionary1 and returning it"""

    if path is None:
        path = []

    for key in dictionary2:
        if key in dictionary1:
            if isinstance(dictionary1[key], dict) and isinstance(dictionary2[key], dict):
                merge(dictionary1[key], dictionary2[key], path + [str(key)])
            elif dictionary1[key] == dictionary2[key]:
                pass  # same leaf value
            else:
                logger.send_exception("Conflict at %s" % ".".join(path + [str(key)]))
                raise InternalServerError("Failed to merge state machine ticket.")
        else:
            dictionary1[key] = dictionary2[key]

    return dictionary1


def merge_tickets(event: dict) -> dict:
    """this function will merge the state machine ticket when there is a
    parallel state
    """
    merged_list = {}
    logger.send_debug(event["ticket"]["list"])

    for ticket in list(event["ticket"]["list"]):
        merge(merged_list, ticket)

    return merged_list


def convert_case(json_event: dict):
    logger.send_debug(json_event)

    for k, v in json_event.items():
        if k in ["source", "origin", "region", "operation", "status", "environment"]:
            json_event[k] = v.upper()
        elif k in ["org_name", "email"]:
            json_event[k] = v.lower()
        elif k in ["first_name", "last_name"]:
            json_event[k] = v.title()
        elif isinstance(v, list):
            for i in v:
                if isinstance(i, dict):
                    convert_case(i)

    return json_event


def build_ticket(event: dict) -> dict:
    convert_case(event["ticket"]["request_data"])

    execution = new_execution(event).asdict()

    logger.send_debug(execution)

    return execution


# Restructure to a dataclass
def build_sns_contract(event: dict) -> dict:
    org = get_org_data_from_event(event)
    return {
        "type": event.get("operation"),
        "result": event.get("execution_status"),
        "request_urn": event.get("request_urn"),
        "sources": f"{event.get('request_data').get('origin')}:{event.get('request_data').get('external_urn')}",
        "organization": {
            "region": event.get("region"),
            "environment": event.get("environment"),
            "state": "ACTIVE",
            "stage": org["stage"],
            "parent": event.get("request_data").get("parent") or None,
            "contacts": event.get("request_data").get("contacts", []),
            "org_name": org["name"],
            "org_key": org["key"],
            "org_id": org["id"],
            "org_uuid": org["uuid"],
        },
        "entitlements": event.get("final_state").get("entitlements", []),
        "features": event.get("final_state").get("delta", {}),
        "custom": event.get("request_data").get("custom", {}),
        "error-info": event.get("error-info", {}),
    }


# Restructure to a dataclass
def build_sns_message_attributes(event: dict) -> dict:
    org = get_org_data_from_event(event)

    # Limited to 10 attributes
    message_attributes = {
        "origin": {"DataType": "String", "StringValue": event.get("origin")},
        "system": {"DataType": "String", "StringValue": event.get("source")},
        "external_id": {
            "DataType": "String",
            "StringValue": event.get("request_data").get("external_urn"),
        },
        "environment": {"DataType": "String", "StringValue": event.get("environment")},
        "operation": {"DataType": "String", "StringValue": event["operation"]},
        "region": {"DataType": "String", "StringValue": event.get("region")},
        "status": {"DataType": "String", "StringValue": event.get("execution_status")},
        "organization": {
            "DataType": "String.Array",
            "StringValue": json.dumps(
                {
                    "org_name": org["name"],
                    "org_key": org["key"],
                    "org_id": org["id"],
                    "org_uuid": org["uuid"],
                    "stage": org["stage"],
                }
            ),
        },
    }

    if entitlements := [f"{ent['system']}:{ent['product_id']}" for ent in event.get("final_state").get("entitlements", [])]:
        message_attributes["entitlements"] = {
            "DataType": "String.Array",
            "StringValue": json.dumps(entitlements),
        }

    if features := [feat["feature_id"] for feat in event.get("final_state").get("features", [])]:
        message_attributes["features"] = {
            "DataType": "String.Array",
            "StringValue": json.dumps(features),
        }

    logger.send_info(message_attributes)

    return message_attributes


# Restructure to a dataclass
def build_csp_attributes(event: dict) -> dict:
    org = get_org_data_from_event(event)

    attributes = {
        "environment": event.get("environment"),
        "external_id": event.get("request_data").get("external_urn"),
        "org_key": org["key"] or "",
        "operation": event.get("operation"),
        "origin": event.get("origin"),
        "region": event.get("region"),
        "status": event.get("execution_status"),
        "system": event.get("source"),
        # "stage": org["stage"], # # This fails, Attribute key stage not found for message type cbc-customer-lifecycle-event
    }
    if entitlements := [f"{ent['system']}:{ent['product_id']}" for ent in event.get("final_state").get("entitlements", [])]:
        attributes["entitlements"] = entitlements

    if features := [feat["feature_id"] for feat in event.get("final_state").get("features", [])]:
        attributes["features"] = features

    return attributes


def publish_sns(event: dict) -> dict:
    csr = CSR(event["request_data"]["environment"])
    org = get_org_data_from_event(event)

    if org["key"]:
        event["final_state"]["entitlements"] = csr.get_org_entitlements(org["key"])

    logger.send_debug(event)

    if is_test_env():
        internal_ingress_hostname = environ.get("INTERNAL_INGRESS_HOSTNAME")
        sns = boto3.client(
            "sns",
            region_name=environ.get("AWS_REGION", "us-east-1"),
            endpoint_url=f"http://{internal_ingress_hostname}/sns/publish",
        )
    else:
        sns = boto3.client("sns", region_name=environ.get("AWS_REGION", "us-east-1"))

    result = sns.publish(
        TopicArn=environ.get("SNS_TOPIC"),
        Message=json.dumps(
            {
                "default": json.dumps(build_sns_contract(event)),
                "lambda": json.dumps({"ticket": event}),
            }
        ),
        MessageStructure="json",
        MessageAttributes=build_sns_message_attributes(event),
        Subject=event["execution_status"].title(),
    )

    event["final_state"].pop("entitlements", None)

    logger.send_info(build_sns_contract(event))

    return result


def publish_csp(event: dict) -> Union[dict, str]:
    csr = CSR(event["request_data"]["environment"])
    org = get_org_data_from_event(event)

    if org["key"]:
        event["final_state"]["entitlements"] = csr.get_org_entitlements(org["key"])

    logger.send_debug(event)

    if not environ.get("CSP_PUBLISHER_ID"):
        logger.send_info("CSP Publisher ID not defined, skipping publish_csp")
        return {}

    csp_message = {
        "publisherID": environ.get("CSP_PUBLISHER_ID"),
        "messageType": "cbc-customer-lifecycle-event",
        "message": build_sns_contract(event),
        "attributes": build_csp_attributes(event),
    }
    logger.send_info(f"Publishing message to CSP: {csp_message}")
    csp_handler = csp.CSP(event.get("request_urn"), "publish csp")
    _ = csp_handler.publish_csp_message(csp_message)
    logger.send_info(f"Message response from CSP: {_}")

    event["final_state"].pop("entitlements", None)

    return _


def parse_ticket_error(error_object: dict) -> dict:
    print("Parsing Ticket Error", error_object)

    # list of params to be exluded from err details
    exclude_err_params = ["password"]

    # list of params to be excluded from err msg attrs
    exclude_err_attrs = ["password", "response"]

    err = error_object.get("Error")
    if not err:
        return {}

    try:
        # Error cause is json we will parse it to find the error msg
        err_cause = json.loads(error_object["Cause"])
    except json.decoder.JSONDecodeError:
        # Error cause is string return it as error message
        err_msg = error_object["Cause"]
        return {"Error": err, "Message": err_msg}

    # Error cause is json, try to parse errorMessage key
    try:
        # errorMessage key is json, try to extract message key from it
        err_params = json.loads(err_cause["errorMessage"])
        err_details = dict(err_params)

        # exclude params from err details
        for param in exclude_err_params:
            err_details.pop(param, None)

        # exclude params from err_attrs
        for param in exclude_err_attrs:
            err_params.pop(param, None)

        err_msg = err_params.pop("message", "Unable to determine the error message")
        err_attrs = ", ".join(f"{k}: {err_details[k]}" for k in sorted(err_params.keys()))

        if err_attrs:
            err_msg = f"{err_msg} [{err_attrs}]"
    except json.decoder.JSONDecodeError:
        # errorMassage key is string, use it as error message
        err_msg = err_cause["errorMessage"]
        err_details = {"error_code": err.upper(), "message": err_msg}

    return {"Error": err, "Message": err_msg, "Details": err_details}


# Need to find a cleaner solution for this
def finalize_ticket_attributes(event: dict) -> None:
    event["ticket"]["execution_status"] = "FAILED" if event["ticket"]["error-info"] else "SUCCESS"
    event["ticket"]["end_time"] = (datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f"))[:-3] + "Z"
    event.get("ticket", {}).get("request_data", {}).pop("user_token", None)
